theory QImp2 imports Main begin

text{* Implementation of Queues *}

text{* 
  The "standard" functional implementation of queues with two lists.
  Efficient, but does not satisfy the axioms literally because of
  observational equivalence.
  *}

datatype 'a q = Q "'a list" "'a list"

text{* The queue datatype satisfies the following invariant: *}

definition q_inv :: "'a q \<Rightarrow> bool"
where "q_inv q = (case q of Q f b \<Rightarrow> (List.null f \<longrightarrow> List.null b))"

fun q_inv' :: "'a q \<Rightarrow> bool"
where "q_inv' (Q f b) = (List.null f \<longrightarrow> List.null b)"

lemma q_inv'_tp: "\<lbrakk> q_inv' (Q f b); ~ List.null b \<rbrakk> \<Longrightarrow> ~ List.null f"
  apply (induction "f")
  apply (force)+
  done

text{* This function ensures the invariant. *}

fun check :: "'a list\<Rightarrow> 'a list\<Rightarrow> 'a q"
where "check [] ys = Q (rev ys) []"
    | "check xs ys = Q xs ys"

lemma q_inv_check: "q_inv' (check f b)"
  apply (induction f)
  apply (simp_all add: null_def)
  done

text{*
  We can now define the queue operations.
  *}

definition mtQ :: "'a q"
where "mtQ == Q [] []"

fun isEmpty :: "'a q=> bool"
where "isEmpty (Q [] _) =  True"
    | "isEmpty (Q _ _) = False"

fun enq     :: "'a => 'a q => 'a q"
where "enq x (Q f b)= check f (x# b)"

fun front   :: "'a q => 'a"
where "front (Q (f#fs) _)= f" 
    | "front (Q [] _) = hd []"

fun deq     :: "'a q => 'a q"
where "deq (Q (f#fs) b) = check fs b"
    | "deq (Q [] b)     = Q [] []"

text{*
  Here, we show all queue operations satisfy the invariant. 
  *}

lemma q_inv_mt: "q_inv' mtQ"
  by (simp add: mtQ_def)

lemma q_inv_enq: "q_inv' (enq x q)"
  by (induction q, simp add: q_inv_check)

lemma q_inv_deq: "q_inv' (deq q)"
  apply (induction q)
  apply (induct_tac list1)
  apply (simp_all add: q_inv_check)
  done

text{*
  The content of a queue, and equality of queues. 
  *}

fun cont    :: "'a q => 'a list"
where "cont (Q f b)= (f @ rev b)"

definition eq :: "'a q \<Rightarrow> 'a q \<Rightarrow> bool"
where "eq q1 q2 \<equiv> (cont q1 = cont q2)"

lemma mt: "isEmpty (Q f b) \<Longrightarrow> List.null f" by (induction f, simp_all add: null_def)
lemma not_mt: "(\<not> isEmpty (Q f b)) = (\<not> List.null f)" by (induction f, simp_all add: null_def)
lemma cont_mt: "q_inv' q \<Longrightarrow> isEmpty q = List.null (cont q)"
  apply (induction q)
  apply (case_tac "list1") 
  apply (simp_all add: null_def)
  done

text{* This expresses the operations in terms of the content. *}

lemma cont_check: "cont (check f b) = f @ rev b"
  by (induction f, simp_all) 
  
lemma cont_enq: "cont (enq x q) = cont q @ [x]"
  by (induction q, simp_all add: cont_check)

lemma cont_deq: "q_inv' q \<Longrightarrow> cont (deq q) = tl (cont q)"
  apply (induction q)
  apply (case_tac list1)
  apply (auto simp add: cont_check null_def)
  done

lemma cont_front: "q_inv' q \<Longrightarrow> front q = hd (cont q)"
  apply (induction q) 
  apply (case_tac list1) 
  apply (auto simp add: null_def)
  done

text{* We need two small lemmas about lists. *}

lemma hd_cat: "~ List.null xs \<Longrightarrow> hd (xs @ ys) = hd xs"
  by (simp add: null_def)

lemma tl_cat: "~ List.null xs \<Longrightarrow> tl (xs @ ys) = tl xs @ ys" by (simp add: null_def)

text{* We can now prove the datatypes axioms. *}

lemma empty: "isEmpty mtQ"
   by (simp add: mtQ_def)

lemma notEmpty: "q_inv' q \<Longrightarrow> \<not> isEmpty (enq a q)"
  by (simp add: cont_mt q_inv_enq cont_enq null_def)

lemma front_mt: "front (enq a mtQ) = a"
  by (simp add: mtQ_def)

lemma front_nonmt: "\<lbrakk> q_inv' q; \<not> isEmpty q \<rbrakk> \<Longrightarrow> front (enq a q) = front q"
  by (simp add: cont_front q_inv_enq cont_enq hd_cat cont_mt)

lemma deq_mt: "eq (deq (enq a mtQ)) mtQ"
  by (simp add: mtQ_def eq_def)


lemma deq_nonmt: "\<lbrakk> q_inv' q; \<not> isEmpty q \<rbrakk> \<Longrightarrow> eq (deq (enq a q)) (enq a (deq q))"
  apply (induction q)
  apply (simp_all add: q_inv_check eq_def cont_deq cont_enq cont_check tl_cat not_mt)
  done


end
